package auth

import (
	"crypto/x509"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"strings"
	"sync"
	"time"

	"github.com/craigfurman/herottp"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type BearerAuthHeaderBuilder struct {
	uaaURL            string
	clientID          string
	clientSecret      string
	httpClient        boshclient.HTTPClient
	cachedToken       string
	cachedTokenExpiry time.Time
	logger            *log.Logger
	tokenLock         *sync.Mutex
}

type ObtainTokenResponse struct {
	Token            string `json:"access_token"`
	ExpiresInSeconds int    `json:"expires_in"`
}

var MinimumRemainingValidity time.Duration = time.Second * 10

func NewBearerAuthHeaderBuilder(uaaURL, clientID, clientSecret string, disableSSLCertVerification bool, trustedCertPEM []byte, logger *log.Logger) *BearerAuthHeaderBuilder {
	rootCAs := x509.NewCertPool()
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	return &BearerAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		httpClient: herottp.New(herottp.Config{
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
		logger:    logger,
		tokenLock: new(sync.Mutex),
	}
}

func (hb *BearerAuthHeaderBuilder) Build() (string, error) {
	hb.tokenLock.Lock()
	defer hb.tokenLock.Unlock()

	if hb.tokenExpired() {
		hb.logger.Println("no valid UAA token found in cache, obtaining a new one")
		var err error
		var expiresInSeconds int
		hb.cachedToken, expiresInSeconds, err = hb.obtainToken()
		if err != nil {
			return "", err
		}
		hb.cachedTokenExpiry = time.Now().Add(time.Second * time.Duration(expiresInSeconds)).Add(MinimumRemainingValidity * -1)
	} else {
		hb.logger.Println("a valid UAA token was found in cache, will not obtain a new one")
	}
	return fmt.Sprintf("Bearer %s", hb.cachedToken), nil
}

func (hb *BearerAuthHeaderBuilder) obtainToken() (string, int, error) {
	request, err := http.NewRequest(http.MethodPost, fmt.Sprintf("%s/oauth/token", hb.uaaURL), strings.NewReader("grant_type=client_credentials"))
	if err != nil {
		return "", 0, err
	}
	request.Header.Set("Content-type", "application/x-www-form-urlencoded")
	request.SetBasicAuth(hb.clientID, hb.clientSecret)

	response, err := hb.httpClient.Do(request)
	if err != nil {
		return "", 0, fmt.Errorf("Error reaching UAA: %s. Please ensure that properties.<broker-job>.bosh.authentication.uaa.url is correct and reachable.", err)
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		bodyBytes, err := ioutil.ReadAll(response.Body)
		if err != nil {
			bodyBytes = []byte("<error reading body>")
		}
		return "", 0, fmt.Errorf("Error authenticating (%d): %s. Please ensure that properties.<broker-job>.bosh.authentication.uaa is correct and try again.", response.StatusCode, bodyBytes)
	}

	var responseContent ObtainTokenResponse
	if err := json.NewDecoder(response.Body).Decode(&responseContent); err != nil {
		return "", 0, err
	}

	if responseContent.Token == "" {
		return "", 0, fmt.Errorf("no access token in grant %#v", responseContent)
	}

	hb.logger.Printf("obtained UAA token, expires in %d seconds\n", responseContent.ExpiresInSeconds)

	return responseContent.Token, responseContent.ExpiresInSeconds, nil
}

func (hb *BearerAuthHeaderBuilder) tokenExpired() bool {
	if hb.cachedToken == "" {
		return true
	}

	return hb.cachedTokenExpiry.Before(time.Now())
}
